# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu
class fitting_parameters(object):
    """Initializes fitting paramters."""

    def __init__(
        self,
        file_names,
        source,
        distance,
        grid,
        n,
        model_wavelength_grid,
        full_model_grid,
        full_outputs,
        min_wavelength,
        max_wavelength,
        bayesian_fit,
        testing,
    ):
        """Initializes fitting paramters.

        Parameters
        ----------
        file_names : array of strings
            Array of the locations of target files.
            e.g.) array(['desk/put_target_data_here/MSX_LMC_807.csv'])
        source : str
            input source string.
            e.g.) 'desk/put_target_data_here/'
        distance : float
            Distance to source in kpc.
        grid : str
            Name of model grid to be used.
        n : int
            Number of times to scale model grid between config.fitting.lum_min
            and config.fitting.lum_min. A value to alter the density of the grid.
        model_wavelength_grid : array
            A 1-D np.array of the wavelengths of the model grid.
        full_model_grid : astropy table
            1-Column astropy table with arrays of model grid fluxes (wm-2) as rows.
        full_outputs : astropy table
            Model outputs as 2D astropy table.
        min_wavelength : float
            Minimum wavelength to fit (um).
        max_wavelength : float
            Maximum wavelength to fit (um).
        bayesian_fit : bool
            Flag for baysian fitting.
        testing : bool
            Flag for testing, used to limit size of model grid.

        Returns
        -------
        class variable
            Class with all fitting parameters.

        """
        # Initialize user-inputted data
        self.file_names = file_names
        self.source = source
        self.distance = distance
        self.grid = grid
        self.n = n
        self.model_wavelength_grid = model_wavelength_grid
        self.full_model_grid = full_model_grid
        self.full_outputs = full_outputs
        self.min_wavelength = min_wavelength
        self.max_wavelength = max_wavelength
        self.bayesian_fit = bayesian_fit
        self.testing = testing



# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu
# from time import time
import ipdb

from multiprocessing import Process, Value, Manager, Pool, cpu_count
from functools import partial
import numpy as np
from desk.set_up import (
    get_inputs,
    get_data,
    create_output_files,
    get_models,
    config,
    create_full_grid,
)
from desk.fitting import dusty_fit
from desk.outputs import plotting_seds

desk_path = str(__file__.replace("console_commands.py", ""))


def grids():
    # Prints the model grids available for fitting.
    print("\nGrids:")
    for item in config.grids:
        print("\t" + str(item))
    print("\n")


def single_fig():
    plotting_seds.single_figures()


def fit(
    source=desk_path + "put_target_data_here",
    distance=config.fitting["default_distance"],
    grid=config.fitting["default_grid"],
    n=config.fitting["default_number_of_times_to_scale_models"],
    min_wavelength=config.fitting["default_wavelength_min"],
    max_wavelength=config.fitting["default_wavelength_max"],
    multiprocessing=True,
    testing=False,
):
    """
    Fits the seds of sources with specified grid.

    Parameters
    ----------
    source : str
        Name of target in array of strings (or one string).
    distance : float
        Distance to source(s) in kiloparsecs.
    grid : str
        Name of model grid.
    """

    # Set-up ###################################################################
    bayesian_fit = False

    # get data filenames
    file_names = get_data.compile_data(source)

    # remove old / create new output files
    create_output_files.remove_old_output_files()
    create_output_files.make_output_files_dusty()

    # gets models
    grid_dusty, grid_outputs = get_models.get_model_grid(grid, testing)

    # does not scale nanni or grams models
    if grid in config.nanni_grids:
        full_model_grid = grid_dusty
        full_outputs = grid_outputs
        full_outputs.rename_columns(
            ["L", "vexp", "dmdt_ir"], ["lum", "scaled_vexp", "scaled_mdot"]
        )

    # elif grid in config.grams_grids:
    #     full_model_grid = grid_dusty
    #     full_outputs = grid_outputs
    #     full_model_grid.rename_columns(["LSPEC"], ["col0"])
    #     full_outputs.rename_columns(["mdot"], ["scaled_mdot"])
    #
    #     # does not calculate expansion velocity so set as 0
    #     full_outputs.add_column(Column([0] * len(full_outputs), name="scaled_vexp"))

    else:
        # update ids to number in grid
        grid_outputs["number"] = np.arange(0, len(grid_outputs))

        # create scaling factors for larger model grid
        scaling_factors = create_full_grid.generate_scaling_factors(distance, int(n))

        # create larger grid by scaling and appending current grid
        full_outputs = create_full_grid.create_full_outputs(
            grid_outputs, distance, scaling_factors
        )
        full_model_grid = create_full_grid.create_full_model_grid(
            grid_dusty, scaling_factors
        )
        full_outputs.remove_columns(["vexp", "mdot"])

    # get model wavelengths
    model_wavelength_grid = grid_dusty["col0"][0]

    # initialize fitting parameters
    fit_params = get_inputs.fitting_parameters(
        file_names,
        source,
        distance,
        grid,
        n,
        model_wavelength_grid,
        full_model_grid,
        full_outputs,
        min_wavelength,
        max_wavelength,
        bayesian_fit,
        testing,
    )

    if multiprocessing == True:
        # Multi-core fitting
        pool = Pool(processes=cpu_count() - 1)
        mapfunc = partial(dusty_fit.fit_single_source, fit_params=fit_params)
        pool.map(mapfunc, range(len(file_names)))
    else:
        # Single-core fitting
        [dusty_fit.fit_single_source(x, fit_params) for x in range(len(file_names))]

    # creates sed figure
    plotting_seds.create_fig()


def sed():
    plotting_seds.create_fig()


if __name__ == "__main__":
    fit()



# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu

target = dict(distance_in_kpc=50, assumed_gas_to_dust_ratio=200, input_unit="Jy")

path = str(__file__.replace("set_up/config.py", ""))

fitting = dict(
    model_grid="Zubko-Crich-bb",
    lum_min=1000,
    lum_max=150000,
    default_wavelength_min=3.5,
    default_wavelength_max=1000,
    default_distance=50,
    default_number_of_times_to_scale_models=100,
    default_grid="oxygen",
)
output = dict(
    printed_output="True",
    output_unit="Wm^-2",
    # output_unit='Jy',
    create_figure="yes",
)
grids = [
    "Crystalline-20-bb",
    # "H11-LMC",
    # "H11-SMC",
    # "J1000-LMC",
    # "J1000-SMC",
    "Oss-Orich-aringer",
    "Oss-Orich-bb",
    "Zubko-Crich-aringer",
    "Zubko-Crich-bb",
    "arnold-palmer",
    "big-grains",
    "corundum-20-bb",
    "fifth-iron",
    "half-iron",
    "one-fifth-carbon",
    # "grams-oxygen",
    # "grams-carbon",
]
nanni_grids = []
# nanni_grids = ["H11-LMC", "H11-SMC", "J1000-LMC", "J1000-SMC"]
# grams_grids = ["grams-oxygen", "grams-carbon"]



# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu
import os


def remove_old_output_files():
    if os.path.exists("fitting_results.txt"):
        os.remove("fitting_results.txt")


def make_output_files_dusty():
    remove_old_output_files()
    with open("fitting_results.csv", "w") as f:
        f.write("source,grid,teff,tinner,model_id,odep,norm,L,vexp,mdot,file_name\n")
        f.close()



# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu
import os
import sys
import ipdb
import numpy as np
import h5py
import urllib
from astropy.table import Table
from desk.set_up import config


def read_hdf5(filename, testing):
    with h5py.File(filename, "r") as f:
        key = list(f.keys())[0]
        if testing == True:
            data = list(f[key][1:4])
        else:
            data = list(f[key])
    out = Table(np.array(data))
    return out


def get_remote_models(model_grid_name):
    def reporthook(blocknum, blocksize, totalsize):
        readsofar = blocknum * blocksize
        if totalsize > 0:
            percent = readsofar * 1e2 / totalsize
            s = "\r%5.0f%% %*d / %d KB" % (
                percent,
                len(str(totalsize)),
                readsofar / 1e3,
                totalsize / 1e3,
            )
            sys.stderr.write(s)
            if readsofar >= totalsize:  # near the end
                sys.stderr.write("\n")
        else:  # total size is unknown
            sys.stderr.write("read %d\n" % (readsofar,))

    models = {
        "arnold-palmer": "https://stsci.box.com/shared/static/5uw23xy6dzrjeb012tw8250r6zfquq1d.hdf5",
        "big-grains": "https://stsci.box.com/shared/static/iwd1wi62rosqhxps2m5ebl8jmv7edx1z.hdf5",
        "corundum-20-bb": "https://stsci.box.com/shared/static/y4ayifjft106j9qbnrc1nl7gly9ajsof.hdf5",
        "Crystalline-20-bb": "https://stsci.box.com/shared/static/9dnzbkhlfvwsfoocelkfop49wpo3ahig.hdf5",
        "fifth-iron": "https://stsci.box.com/shared/static/xtqkm8htmr3xe1n0mcl2p27smtan93a2.hdf5",
        "grams-carbon": "https://stsci.box.com/shared/static/bb8em53lug4a3alggndcqro6cxxv4rkk.hdf5",
        "grams-oxygen": "https://stsci.box.com/shared/static/12oqk52b52sghzo9zb6smxcvwm0zsxqf.hdf5",
        "H11-LMC": "https://stsci.box.com/shared/static/hgnmd8jwem3pm3ewjazdpssu9fst60hl.hdf5",
        "H11-SMC": "https://stsci.box.com/shared/static/ofes4gsscf0y8g8fu72alu5cx40lry58.hdf5",
        "half-iron": "https://stsci.box.com/shared/static/q5j9ey5ci5vtu2nnl6qf76entxt3xsr5.hdf5",
        "J1000-LMC": "https://stsci.box.com/shared/static/3ed8e7ulpu30cbyri3ex7m08g5jztk6m.hdf5",
        "J1000-SMC": "https://stsci.box.com/shared/static/a7j1z18sz4mr3k2dxqj5cob6wqutfcad.hdf5",
        "one-fifth-carbon": "https://stsci.box.com/shared/static/olpt7phvjd4kin6wn0w9y1wez9037f73.hdf5",
        "Oss-Orich-aringer": "https://stsci.box.com/shared/static/xytr8t139h8zuanwtf773a8bzt32bxcc.hdf5",
        "Oss-Orich-bb": "https://stsci.box.com/shared/static/arczjrc25xi601rlwxco08b14vabpp52.hdf5",
        "Zubko-Crich-aringer": "https://stsci.box.com/shared/static/yq8vbohgybuo7nmkwu0w535q6tjeyw1v.hdf5",
        "Zubko-Crich-bb": "https://stsci.box.com/shared/static/4ocskl59q8tvrhxbpcvpo26qm1wy9ytg.hdf5",
    }
    outputs = {
        "arnold-palmer": "https://stsci.box.com/shared/static/bc8up4ra3b1a8crpebnjvhl9bxnhws52.hdf5",
        "big-grains": "https://stsci.box.com/shared/static/apiwiek17yjjey6hx7b87w5oeupnxbq7.hdf5",
        "corundum-20-bb": "https://stsci.box.com/shared/static/o6bx04qfbs4nof93e16jr0gl0lt04vmi.hdf5",
        "Crystalline-20-bb": "https://stsci.box.com/shared/static/aee6evhqfp6m8p8wsqnj6finsxj022bq.hdf5",
        "fifth-iron": "https://stsci.box.com/shared/static/4jkhznyw1566wprc4w3og0ps2e6pyk60.hdf5",
        "grams-carbon": "https://stsci.box.com/shared/static/m0gc3eils06b0zcy1t75l8x6h8ftpij1.hdf5",
        "grams-oxygen": "https://stsci.box.com/shared/static/mc6h7hhew918du9nudez1w2obja1ydqy.hdf5",
        "H11-LMC": "https://stsci.box.com/shared/static/ndzqw7s6yr8kmmkq71hnvi10gb42fx6k.hdf5",
        "H11-SMC": "https://stsci.box.com/shared/static/9q7yh6mawezxbt5ipvmqrui2r0gld4f0.hdf5",
        "half-iron": "https://stsci.box.com/shared/static/r26hehfxy92cb5cweihkl9159xr9a7fe.hdf5",
        "J1000-LMC": "https://stsci.box.com/shared/static/242covljjj7iri59tjsd9ah80rbgsptg.hdf5",
        "J1000-SMC": "https://stsci.box.com/shared/static/ampoqu2a9azpmrql09b51lv8rch2ubmb.hdf5",
        "one-fifth-carbon": "https://stsci.box.com/shared/static/vfg9ab4gs1y5zsvmrd8qd1g3xhfno4nz.hdf5",
        "Oss-Orich-aringer": "https://stsci.box.com/shared/static/t75dvs50380lpffsjh2f4yjrvgoi8r4y.hdf5",
        "Oss-Orich-bb": "https://stsci.box.com/shared/static/ov83sjmx69ddgy8bahg5p8yu2c2gxfqo.hdf5",
        "Zubko-Crich-aringer": "https://stsci.box.com/shared/static/0nrjesla9jgxh8ioatnmrmrc5tkcf3pf.hdf5",
        "Zubko-Crich-bb": "https://stsci.box.com/shared/static/d0651w7ztavoiir7341lda8xzbwrfq9g.hdf5",
    }
    # grid_csv_link = [
    #     "https://stsci.box.com/shared/static/jxtabj6h5zi7a8ggb0mbtag3v6b4b9i7.csv",
    #     "https://stsci.box.com/shared/static/z0hnvo0hz0qxcwlcvbmf01avdjlhewah.csv",
    #     "https://stsci.box.com/shared/static/j7f01dwbsbluq4htdwlsx3v0jknye883.csv",
    #     "https://stsci.box.com/shared/static/o6dp1cyj86jw2ifuuco983kgb2uf16sf.csv",
    #     "https://stsci.box.com/shared/static/fuhrcgo4vh4m7ca121p1f0vuss4vszcz.csv",
    #     "https://stsci.box.com/shared/static/he92pt1yov18x6ska5f4f4q94gal6tr5.csv",
    #     "https://stsci.box.com/shared/static/pnljezbu8c7rcyorb27yh6hylk5wy2fs.csv",
    #     "https://stsci.box.com/shared/static/x9fg8lacbbiye591tz723ilcqgcomyjh.csv",
    #     "https://stsci.box.com/shared/static/4b9rc9zl110khexeuo8thb8r58mrl73g.csv",
    #     "https://stsci.box.com/shared/static/4gxmapkhml7lnxsiu9dzto8a7zwnmzd5.csv",
    #     "https://stsci.box.com/shared/static/kfaq7jqt4t69vitf1cifk5gurx7lnogn.csv",
    #     "https://stsci.box.com/shared/static/xf6rrfnp1jdin94iwhwb47cxoplu3aux.csv",
    #     "https://stsci.box.com/shared/static/5zqshpzw748n0doykm7d5cxnlvhue4kk.csv",
    #     "https://stsci.box.com/shared/static/x7328lqy8fyg19wswqwq4md4gm398ewi.csv",
    #     "https://stsci.box.com/shared/static/8j73qttzj10eu5sct0ds0835uxbxu3s8.csv",
    #     "https://stsci.box.com/shared/static/ldocm4yto4vtxfcg21o23a8uilb3a3do.csv",
    #     "https://stsci.box.com/shared/static/iy6dsqrnhi96ecnlsozo1ljr5g8b1jwx.csv",
    # ]
    # grid_fits_link = [
    #     "https://stsci.box.com/shared/static/1suiv6nqbc7yoqutli2q9q98ste31o1y.fits",
    #     "https://stsci.box.com/shared/static/bq384o6m1xm41bkx7s8oloohqkd87hy8.fits",
    #     "https://stsci.box.com/shared/static/faekj3usdme9go3lmynga3uk0oln8v3k.fits",
    #     "https://stsci.box.com/shared/static/w0e3f1a0r886u17vjosk8lvqrnvnw6ce.fits",
    #     "https://stsci.box.com/shared/static/e0nvigaed722a67043rvmqhjzkonttp4.fits",
    #     "https://stsci.box.com/shared/static/60sns7ua91ixuz4rddul2ufrmc0pu5yo.fits",
    #     "https://stsci.box.com/shared/static/hr113z7lgvggyybh9eygi6nj69gu1gzv.fits",
    #     "https://stsci.box.com/shared/static/hvgfnug5xxcepcz083cnrizsukgddttw.fits",
    #     "https://stsci.box.com/shared/static/c5gdw6o3b96kaphe07y459zd7ba6590f.fits",
    #     "https://stsci.box.com/shared/static/b78ks7cjzenagoznqqrdfoyh6bonw2sy.fits",
    #     "https://stsci.box.com/shared/static/lw6gp2s8surgbbtmcnyev673faxqg9h6.fits",
    #     "https://stsci.box.com/shared/static/yq6pbvepyt8a420ealg6i111xc7hcty5.fits",
    #     "https://stsci.box.com/shared/static/qjuq780xr0ihj9p909wmijwkpclu2njd.fits",
    #     "https://stsci.box.com/shared/static/ht6edjsrupwhytwuwa9hjg53dof2zzy0.fits",
    #     "https://stsci.box.com/shared/static/n1ng9f4s8s7ps0ah8vp24gk0fzsxp0aj.fits",
    #     "https://stsci.box.com/shared/static/l1qayndbz40lwewvlpg1qxu22rs9f9xu.fits",
    #     "https://stsci.box.com/shared/static/jho46vam6d0k9jwg78gxx4eh7emwx4oz.fits",
    # ]
    if model_grid_name in models:
        url_outputs = outputs[model_grid_name]
        url_models = models[model_grid_name]
    else:
        raise ValueError(
            "ERROR: Model name not an option. \nCurrent downloadable options: \n \t Zubko-Crich-aringer \n \t Oss-Orich-bb \n \t Oss-Orich-aringer \n \t Crystalline-20-bb \n \t corundum-20-bb \n \t arnold-palmer \n \t big-grains \n \t fifth-iron \n \t one-fifth-carbon"
        )

    # \n Padova options: J400, J1000, H11, R12, R13'

    # Download files
    print(". . . Downloading model: " + model_grid_name + " . . .")
    urllib.request.urlretrieve(
        url_outputs, config.path + "/models/" + model_grid_name + "_outputs.hdf5"
    )
    urllib.request.urlretrieve(
        url_models,
        config.path + "/models/" + model_grid_name + "_models.hdf5",
        reporthook,
    )
    print("Download Complete!")


def check_models(model_grid):

    """
    Checks if model grids are available and returns the full path to the model.
    If the model is not downloaded, it is downloaded via Box.

    Parameters
    ----------
    model_grid : str
        Name of model grid to use.

    Returns
    -------
    outputs_file: str
        The full path/name of the model outputs file.
    models_file: str
        The full path/name of the model grid file.

    """

    outputs_file = config.path + "models/" + model_grid + "_outputs.hdf5"
    models_file = config.path + "models/" + model_grid + "_models.hdf5"

    # Checks if grid is available
    if os.path.isfile(outputs_file) and os.path.isfile(models_file):
        print("\nYou already have the grid!\n")
    else:
        # asks if you want to download the models
        print("Models not found locally")
        get_remote_models(model_grid)
    return (outputs_file, models_file)


def get_model_grid(grid, testing):
    """
    Gets the real model grid name if the defaults were chosen,and runs check_models.

    Parameters
    ----------
    grid : str
        Model grid name.

    Returns
    -------
    grid_dusty : 2 column astropy table with array of wavelengths and array of
    fluxes in each column of each row
        The (intial) model grid wavelengths and fluxes. This is not the full model
        grid with appended scaled models.

    grid_outputs : astropy table
        The model grid parameters corresponding to the grid_dusty model grids
    """

    # User input for models
    if grid == "carbon":
        model_grid = "Zubko-Crich-bb"
    elif grid == "oxygen":
        model_grid = "Oss-Orich-bb"
    else:
        if grid in config.grids:
            model_grid = grid
        else:
            raise ValueError(
                "\n\nUnknown grid. Please make another model selection.\n\n To see options use: desk grids\n"
            )
    outputs_file_name, models_file_name = check_models(model_grid)

    grid_dusty = read_hdf5(models_file_name, testing)
    grid_outputs = read_hdf5(outputs_file_name, testing)

    return grid_dusty, grid_outputs



# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu
# Script adapted from the BEAST (https://github.com/BEAST-Fitting)
import ipdb
import numpy as np
import math

from desk.outputs import plot_pdf


class pdf1d:
    def __init__(self, gridvals, nbins, logspacing=False, minval=None, maxval=None):
        """
        Create an object which can be used to efficiently generate a 1D pdf for an observed object.

        Parameters
        ----------

        gridvals: array-like
            values of the quantity for all the grid points

        nbins: int
            number of bins to use for the 1D pdf

        logspacing: bool
            whether to use logarithmic spacing for the bins

        minval, maxval: float (optional)
            override the range for the bins. this can be useful to make
            sure that the pdfs for different runs have the same bins
        """
        self.nbins = nbins
        self.n_gridvals = len(gridvals)
        self.logspacing = logspacing

        # grab copy of gridvals that can be edited without messing with original
        tgridvals = np.array(gridvals)
        self.n_indxs = len(gridvals)

        if len(tgridvals) <= 0:
            # this is a hack to just get the code to work when
            # all the possible values are negative and the requested
            # pdf is for log x values
            self.bad = True
            self.bin_vals = np.linspace(0.0, 1.0, num=self.nbins)
        else:
            self.bad = False
            # set bin ranges
            self.min_val = tgridvals.min() if minval is None else minval
            self.max_val = tgridvals.max() if maxval is None else maxval

            # if log spacing requested, make the transformation
            if logspacing:
                self.min_val = math.log10(self.min_val)
                self.max_val = math.log10(self.max_val)
                tgridvals = np.log10(tgridvals)

            # set bin widths
            if self.nbins > 1:
                self.bin_delta = (self.max_val - self.min_val) / (self.nbins - 1)
            else:
                self.bin_delta = 1

            # set values for the bin middles/edges
            self.bin_vals = self.min_val + np.arange(self.nbins) * self.bin_delta
            self.bin_edges = (
                self.min_val + (np.arange(self.nbins + 1) - 0.5) * self.bin_delta
            )

            # get PDF bin associated with each grid val
            pdf_bin_num = np.digitize(tgridvals, self.bin_edges)

            # array to hold indices for each bin
            # (like the IDL version returned by the histogram function)
            pdf_bin_indxs = []

            for i in range(nbins):
                # find the indicies for the current bin
                cur_bin_indxs, = np.where(pdf_bin_num == (i + 1))

                # save them
                pdf_bin_indxs.append(cur_bin_indxs)

            # transform the bin edges back to linear spacing if log spacing
            #  was asked for
            if logspacing:
                self.bin_vals = np.power(10.0, self.bin_vals)
                self.bin_edges = np.power(10.0, self.bin_edges)

            self.pdf_bin_indxs = pdf_bin_indxs

    def gen1d(self, gindxs, weights):
        """
        Determines the probabilities for each bin in weights.

        Parameters
        ----------
        gindxs : array
            Array of indices excluding those with weights below threshold.
        weights : array
            grid_weights * priors * log_liklihoods

        Returns
        -------
        self.bin_vals: array
            parameter bins
        _vals_1d: array
            sums of probabilities for each bin
        """

        if self.bad:
            return (self.bin_vals, np.zeros((self.nbins)))
        else:
            _tgrid = np.zeros(self.n_gridvals)
            _tgrid[gindxs] = weights
            _vals_1d = np.zeros(self.nbins)
            for i in range(self.nbins):
                if len(self.pdf_bin_indxs[i]) > 0:
                    _vals_1d[i] = np.sum(_tgrid[self.pdf_bin_indxs[i]])

            return (self.bin_vals, _vals_1d)


def par_pdf(par, full_outputs, probs):
    # create 1d-pdfs
    pdf = pdf1d(full_outputs[par], 50, logspacing=True)

    # calculates probabilities for each bin (bin width set by dif in bins)
    bins, bin_total_prob = pdf1d.gen1d(pdf, np.arange(0, len(full_outputs)), probs)

    # most likely value
    best = bins[np.argmax(bin_total_prob)]

    # plot probability function
    plot_pdf.plot(par, pdf, bin_total_prob, best)
    return best



# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu


class Fitting_Range_Error(ValueError):
    pass


class BadFilenameError(ValueError):
    pass


class BadSourceDirectoryError(ValueError):
    pass



import ipdb
import numpy as np


class fit:

    """
    Fitting tools for least square fit.
    """

    def trim(data, model_trim):

        """
        Removes data outside of wavelegth range of model grid.

        Parameters
        ----------
        data : 2-D array
            input data in 2-D array of wavelength and flux.
        model_trim : type
            input model in 2-D array of wavelength and flux.

        Returns
        -------
        2-D array
            Trimmed input model in 2-D array of wavelength and flux.

        """
        indexes = np.where(
            np.logical_and(
                model_trim[0] >= np.min(data[0]), model_trim[0] <= np.max(data[0])
            )
        )
        return model_trim[0][indexes], model_trim[1][indexes]

    def find_closest(data_wave, model_wave, model_flux):

        """
        Find model fluxes closest in wavelength to data.

        Parameters
        ----------
        data_wave : 1-D array
            wavelengths of data in microns.
        model_wave : 1-D array
            wavelengths of model in microns.
        model_flux : 1-D array
            scaled flux of model in W M-2

        Returns
        -------
        closest_model_flux: array
            subset of model fluxes

        """
        closest = np.searchsorted(model_wave, data_wave)
        idx = np.clip(closest, 1, len(model_wave) - 1)
        left = model_wave[idx - 1]
        right = model_wave[idx]
        idx -= data_wave - left < right - data_wave
        closest_model_flux = model_flux[idx]
        return closest_model_flux

    def least2_liklihood(_data, _model):
        # least squares fit
        # ipdb.set_trace()
        _stat = np.nansum(np.square(_data - _model) / _model)
        prob = np.exp(-0.5 * np.float128(_stat))
        return prob

    def fit_data(data, model):

        """
        Trims the data, finds the closest match, and returns chi square value.

        Parameters
        ----------
        data : 2D array
            Data with wavelength in microns and flux in W M-2
        model : 2D array
            model with wavelength in microns and scaled flux in W M-2

        Returns
        -------
        stats: float
            chi square value.

        """
        trimmed_model_wave, trimmed_model_flux = fit.trim(data, model)
        matched_model = fit.find_closest(
            data[0], trimmed_model_wave, trimmed_model_flux
        )
        liklihood = fit.least2_liklihood(data[1], matched_model)
        # ipdb.set_trace()
        return liklihood



# Steve Goldman, Space Telescope Science Institute, sgoldman@stsci.edu
import math
import ipdb
from tqdm import tqdm
import numpy as np
from copy import deepcopy
from astropy.table import Table, Column, vstack
from desk.set_up import config

__all__ = ["generate_scaling_factors", "create_full_outputs", "create_full_model_grid"]


def generate_scaling_factors(distance, number_of_models):
    """
    Creates arrays of model fluxes from lum_min to lum_max.

    Parameters
    ----------
    distance : float
        Distance to source in Kpc.

    Returns
    -------
    scaling_vals : 1D numpy array
        An array of scaling factors that the model grid will be scaled to.
        As the model grids are used to create higher-luminosity models through scaling,
        this sets the different luminosities that each grid will have (i.e. the model
        grid will be appended by the same model grid scaled by each of the values
        in scaling_factors)
    """

    distance_norm = math.log10(((float(distance) / 4.8482e-9) ** 2) / 1379)
    scaling_vals = (
        np.linspace(
            np.log10(config.fitting["lum_min"]),
            np.log10(config.fitting["lum_max"]),
            number_of_models,
        )
        - distance_norm
    )
    return scaling_vals


def create_full_outputs(_grid_outputs, distance, trials):

    """
    Returns the input grid for each luminosity in the form of trial.
    (scaling factor not including distance; e.g. -12.5, -13.5) specified by trials.
    Also includeds the scled vexp and mass loss rate.

    Parameters
    ----------
    _grid_outputs : astropy table
        model grid.
    distance : int or float
        distance in kpc.
    trials : array
        Array of scaling factors that does not include distance.

    Returns
    -------
    grid_outputs : astropy table
        The input astropy table but appeded to itself for each scaling factor.
        With the added columns for luminosity, scaled mdot, and sclaed vexp

    """

    grid_template = deepcopy(_grid_outputs)
    distance_norm = math.log10(((float(distance) / 4.8482e-9) ** 2) / 1379)

    print(
        "Scaling to full grid ("
        + "{:,}".format((len(grid_template) * len(trials)))
        + " models)"
    )

    # for each scaling value, create and append a grid
    for i, trial in enumerate(tqdm(trials)):
        appended_trials = Column(np.full(len(grid_template), trial), name="trial")
        if i == 0:
            _grid_outputs.add_column(appended_trials)
        else:
            add_trials = deepcopy(grid_template)
            add_trials.add_column(appended_trials)
            _grid_outputs = vstack((_grid_outputs, add_trials))

    # adds luminosity column
    luminosity = Column(
        np.array(np.power(10.0, distance_norm - _grid_outputs["trial"] * -1)),
        name="lum",
    )
    _grid_outputs.add_column(luminosity.astype(int))

    # adds scaled parameters of gas mass loss rate and expansion velocity
    scaled_vexp = Column(
        np.array(_grid_outputs["vexp"])
        * (np.array(_grid_outputs["lum"]) / 10000) ** 0.25,
        name="scaled_vexp",
    )
    scaled_mdot = Column(
        np.array(_grid_outputs["mdot"])
        * ((np.array(_grid_outputs["lum"]) / 10000) ** 0.75)
        * (config.target["assumed_gas_to_dust_ratio"] / 200) ** 0.5,
        name="scaled_mdot",
    )
    _grid_outputs.add_columns([scaled_vexp, scaled_mdot])
    return _grid_outputs


def create_full_model_grid(grid_dusty, scaling_factors):
    """
    Returns model flux grids for each luminosity scaling (scaling_factors).

    Parameters
    ----------
    grid_dusty : Astropy table with 1 column with flux grid in each row
        The (un-trimmed) flux grid in w/m^2 for each model, and for each luminosity
    scaling_factors : 1D array
        scaling factors.

    Returns
    -------
    type: Astropy table with 1 column with scaled flux grid (w m-2) in each row
        scaled flux grid

    """
    scaled_rows = []
    for val in scaling_factors:
        for row in grid_dusty["col1"]:
            scaled_rows.append(row * np.power(10, val))
    scaled_grid = Table([scaled_rows])
    return scaled_grid
